"use strict";
/**
 * Basic workflow -
 * 1. Connects to comms created by pyflyby
 * 2. Recieves imports added by pyflyby via PYFLYBY_COMMS.MISSING_IMPORTS
 * 3. Sends import statements recived in previous step to kernel for formatting using tidy_imports
 * 4. Recieves formatted imports via PYFLYBY_COMMS.FORMAT_IMPORTS which are added at suitable location in notebook
 *
 * Selecting cell where imports are added -
 * 1. First cell with 'pyflyby-cell' tag, if not  present then next step.
 * 2. First code cell which does not contain magic command
 * 3. If selected cell doesn't contain any import statement, add a new cell above the code cell.
 *
 * Selecting insert location inside the cell -
 * 1. If PYFLYBY_END_MSG is present, import is added above it.
 * 2. Added import after last import statement in code cell. Identifying import statement is
 *    is done by simple heuristics. This step can be shifted to pyflyby where python parser can be used to
 *    determine it accurately.
 */
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
// Lumino imports
const algorithm_1 = require("@lumino/algorithm");
const widgets_1 = require("@lumino/widgets");
const apputils_1 = require("@jupyterlab/apputils");
const settingregistry_1 = require("@jupyterlab/settingregistry");
const debug_1 = require("debug");
const react_1 = __importDefault(require("react"));
// relative imports
const cellUtils_1 = require("./cellUtils");
const constants_1 = require("./constants");
const log = debug_1.debug('PYFLYBY:');
// We'd like to show the notification only once per session, not for each notebook
let _userWasNotified = false;
/**
 * An extension that adds pyflyby integration to a single notebook widget
 */
class PyflyByWidget extends widgets_1.Widget {
    constructor(context, panel, settingRegistry) {
        super();
        this._context = null;
        this._sessionContext = null;
        this._settings = null;
        this._comms = {};
        // get a reference to the settings registry
        settingRegistry.load('@deshaw/jupyterlab-pyflyby:plugin').then((settings) => {
            this._settings = settings;
            const enabled = settings.get('enabled').user || settings.get('enabled').composite;
            if (enabled) {
                this._sessionContext.kernelChanged.connect(this._handleKernelChange, this);
                this._sessionContext.statusChanged.connect(this._handleKernelStatusChange, this);
            }
        }, (err) => {
            log('PYFLYBY extension has been disabled');
        });
        this._context = context;
        this._sessionContext = context.sessionContext;
    }
    async _launchDialog() {
        /**
         * Since we are making the first import, create a new dialog
         */
        const dialog = new apputils_1.Dialog({
            title: 'PYFLYBY',
            body: `PYFLYBY will be adding imports to the first code cell in the notebook.
            To disable the PYFLYBY extension or to disable this notification in future, go
            to Settings -> Advanced Settings Editor and choose PYFLYBY preferences tab`,
            buttons: [apputils_1.Dialog.okButton()],
        });
        try {
            await dialog.launch();
        }
        catch (e) {
            console.error(e);
        }
    }
    /**
     * All the logic related to finding the right cell
     */
    _findImportCoordinates() {
        const { model } = this._context;
        let pyflybyCellIndex = algorithm_1.ArrayExt.findFirstIndex(algorithm_1.toArray(model.cells), (cell, index) => {
            const tags = cell.metadata.get('tags');
            return !!(tags && tags.indexOf(constants_1.PYFLYBY_CELL_TAG) !== -1);
        });
        /**
         * Since the cell doesn't exist, we make one or, if the first
         * code cell contains an import block, put it below that.
         */
        if (pyflybyCellIndex === -1) {
            pyflybyCellIndex = cellUtils_1.findCell(algorithm_1.toArray(model.cells));
        }
        let cell = model.cells.get(pyflybyCellIndex);
        let position = cellUtils_1.findLinePos(model.cells.get(pyflybyCellIndex));
        if (position === -1) {
            cell = this._context.model.contentFactory.createCodeCell({
                cell: {
                    source: `${constants_1.PYFLYBY_START_MSG}\n\n${constants_1.PYFLYBY_END_MSG}`,
                    cell_type: 'code',
                    metadata: {},
                },
            });
            this._context.model.cells.insert(pyflybyCellIndex, cell);
            position = constants_1.PYFLYBY_START_MSG.length + 1;
        }
        cell.metadata.set('tags', [constants_1.PYFLYBY_CELL_TAG]);
        return { cellIndex: pyflybyCellIndex, position };
    }
    /**
     * Adds the import block to the appropriate cell at the appropriate
     * location.
     *
     * @param importBlock - the import statement or block of import statements
     */
    _insertImport({ missing_imports: importBlock }) {
        let p = null;
        if (!_userWasNotified && !this._settings.get('disableNotification').user) {
            p = this._launchDialog();
            _userWasNotified = true;
        }
        else {
            p = Promise.resolve();
        }
        const { model } = this._context;
        // Find the right cell and position to insert cell
        const { position, cellIndex } = this._findImportCoordinates();
        const cell = model.cells.get(cellIndex);
        const insertIndex = position === -1 ? cell.value.text.length : position;
        let toInsert = cell.value.text.length === 0 ? importBlock : `${importBlock}\n`;
        if (insertIndex !== 0 && cell.value.text[insertIndex - 1] !== `\n`) {
            toInsert = `\n${toInsert}`;
        }
        cell.value.insert(insertIndex, toInsert);
        this._sendFormatCodeMsg();
        return p;
    }
    _sendFormatCodeMsg() {
        let pyflybyCellIndex = algorithm_1.ArrayExt.findFirstIndex(algorithm_1.toArray(this._context.model.cells), (cell, index) => {
            const tags = cell.metadata.get('tags');
            return !!(tags && tags.indexOf(constants_1.PYFLYBY_CELL_TAG) !== -1);
        });
        if (pyflybyCellIndex !== -1) {
            const cellSource = this._context.model.cells
                .get(pyflybyCellIndex)
                .toJSON().source;
            const comm = this._comms[constants_1.PYFLYBY_COMMS.FORMAT_IMPORTS];
            if (comm && !comm.isDisposed) {
                comm.send({
                    input_code: cellSource,
                    type: constants_1.PYFLYBY_COMMS.FORMAT_IMPORTS,
                });
            }
        }
    }
    _getCommMsgHandler() {
        return (msg) => {
            const msgContent = msg.content.data;
            switch (msgContent.type) {
                case constants_1.PYFLYBY_COMMS.MISSING_IMPORTS:
                    this._insertImport(msg.content.data).catch(console.error);
                    break;
                case constants_1.PYFLYBY_COMMS.FORMAT_IMPORTS:
                    this._formatImports(msg.content.data);
                    break;
                case constants_1.PYFLYBY_COMMS.INIT:
                    this._initializeComms().catch(console.error);
                    break;
                default:
                    break;
            }
        };
    }
    async _initializeComms() {
        const { kernel } = this._sessionContext.session;
        if (!kernel) {
            return;
        }
        // Open the comm
        const targetName = constants_1.PYFLYBY_COMMS.MISSING_IMPORTS;
        const comm = kernel.createComm(targetName);
        comm.onMsg = this._getCommMsgHandler();
        try {
            comm.open();
        }
        catch (e) {
            console.error(`Unable to open PYFLYBY comm - ${e}`);
        }
        const formatMsgComm = kernel.createComm(constants_1.PYFLYBY_COMMS.FORMAT_IMPORTS);
        formatMsgComm.onMsg = this._getCommMsgHandler();
        formatMsgComm.onClose = (msg) => {
            const commId = msg.content.comm_id;
            delete this._comms[commId];
        };
        this._comms[constants_1.PYFLYBY_COMMS.FORMAT_IMPORTS] = formatMsgComm;
        try {
            formatMsgComm.open();
        }
        catch (e) {
            console.error(`Unable to open PYFLYBY comm - ${e}`);
        }
        kernel.registerCommTarget(constants_1.PYFLYBY_COMMS.INIT, (comm, msg) => {
            comm.onMsg = this._getCommMsgHandler();
        });
        return Promise.resolve();
    }
    _formatImports(msgData) {
        const { formatted_code: formattedCode } = msgData;
        let pyflybyCellIndex = algorithm_1.ArrayExt.findFirstIndex(algorithm_1.toArray(this._context.model.cells), (cell, index) => {
            const tags = cell.metadata.get('tags');
            return !!(tags && tags.indexOf(constants_1.PYFLYBY_CELL_TAG) !== -1);
        });
        if (pyflybyCellIndex !== -1) {
            const cell = this._context.model.cells.get(pyflybyCellIndex);
            cell.value.remove(0, cell.value.text.length);
            cell.value.insert(0, formattedCode);
        }
    }
    async _handleKernelChange(sender, kernelChangedArgs) {
        return await this._initializeComms();
    }
    _handleKernelStatusChange(sender, args) {
        if (args === 'restarting') {
            return this._initializeComms();
        }
    }
}
/**
 * An extension that adds pyflyby integration to a notebook widget
 */
class PyflyByWidgetExtension {
    constructor(settingRegistry) {
        this._settingRegistry = null;
        // get a reference to the settings registry
        // This is shared between all notebooks. I.e. not possible to
        // have different pyflyby settings for different notebooks
        this._settingRegistry = settingRegistry;
        this._loadSettings().catch(console.error);
    }
    async _loadSettings() {
        try {
            await this._settingRegistry.load('@deshaw/jupyterlab-pyflyby:plugin');
            log('Successfully loaded PYFLYBY extension settings');
        }
        catch (e) {
            console.error('Settings could not be loaded');
        }
    }
    createNew(panel, context) {
        return new PyflyByWidget(context, panel, this._settingRegistry);
    }
}
async function isPyflybyInstalled() {
    const response = await fetch('/pyflyby/pyflyby-status');
    const pyflybyStatus = await response.json();
    return pyflybyStatus.status;
}
async function installPyflyby() {
    try {
        await fetch('/pyflyby/install-pyflyby', { method: 'POST' });
    }
    catch (err) {
        const errMsg = await err.json();
        console.error(errMsg.result);
    }
}
async function disableJupyterlabPyflyby(registry) {
    try {
        await fetch('/pyflyby/disable-pyflyby', {
            method: 'POST',
            mode: 'cors',
            cache: 'no-cache',
            credentials: 'include',
            headers: { 'Content-type': 'application/x-www-form-urlencoded' },
            body: new URLSearchParams('installDialogDisplayed=true'),
        });
    }
    catch (err) {
        const errMsg = await err.json();
        console.error(errMsg.result);
    }
    await registry.reload('@deshaw/jupyterlab-pyflyby:plugin');
}
const installationBody = (react_1.default.createElement("div", null,
    react_1.default.createElement("p", null,
        "To use @deshaw/jupyterlab-pyflyby,",
        ' ',
        react_1.default.createElement("a", { href: "https://github.com/deshaw/pyflyby/blob/master/README.rst", style: { color: '#0000EE' }, target: "_blank", rel: "noopener noreferrer" }, "pyflyby"),
        ' ',
        "ipython extension needs to be installed."),
    react_1.default.createElement("br", null),
    react_1.default.createElement("p", null, "Clicking on \"Install\" will run following command"),
    react_1.default.createElement("div", { style: {
            font: 'monospace',
            color: '#ffffff',
            backgroundColor: '#000000',
            marginTop: '5px',
        } }, "$ py pyflyby.install_in_ipython_config_file"),
    react_1.default.createElement("br", null)));
const extension = {
    id: '@deshaw/jupyterlab-pyflyby:plugin',
    autoStart: true,
    requires: [settingregistry_1.ISettingRegistry],
    activate: async function (app, registry) {
        const settings = await registry.load('@deshaw/jupyterlab-pyflyby:plugin');
        const enabled = settings.get('enabled').user || settings.get('enabled').composite;
        const dialogDisplayedEarlier = settings.get('installDialogDisplayed').user;
        if (enabled) {
            const response = await isPyflybyInstalled();
            if (response !== 'loaded') {
                if (dialogDisplayedEarlier) {
                    // Dialog to install pyflyby ipython extensions was displayed earlier,
                    // install it since user is trying to use pyflyby by manually enabling
                    // jupyterlab-pyflyby
                    await installPyflyby();
                }
                else {
                    const result = await apputils_1.showDialog({
                        title: 'Installation required',
                        body: installationBody,
                        buttons: [
                            apputils_1.Dialog.okButton({
                                label: 'Install'
                            }),
                            apputils_1.Dialog.cancelButton({ label: 'Cancel', displayType: 'default' }),
                        ],
                        defaultButton: 0,
                    });
                    result.button.accept
                        ? await installPyflyby()
                        : await disableJupyterlabPyflyby(registry);
                }
            }
        }
        app.docRegistry.addWidgetExtension('Notebook', new PyflyByWidgetExtension(registry));
    },
};
exports.default = extension;
